# -*- Mode: Python; tab-width: 4 -*-
#
#	$Id: elements.py,v 1.4 1999/06/17 17:38:55 pavlos Exp $
#	Author: Pavlos Christoforou <pavlos@gaaros.com>
#
# ======================================================================
#
# This software, its author and the author's donkey are all certified
# as open source. Use and abuse us in any way you see fit.
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY GAAROS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL GAAROS OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
########################################################################

__version__='$Revision: 1.4 $'[11:-2]


"""/nDocument elements

   All elements (with the exception of the basic phrase unit, and
   UL,OL etc) should conform to the following interface:

   - format: formats elements based on the document style

   - more: As elements are responsible for their own formating, this
     method returns two objects, the first is the part of the element
     that fits in the current page and a second containing any
     leftovers.

   - to_PDF: returns a PDF representation of the element.
   
   - get_bbox: returns the bounding box of the object
   
   Text Elements  

     the basic text element is a phrase which can contain one or more
     characters.

     the rest of the text elements correspond loosely to the standard
     HTML elements (H1,H2,H3,P,UL,OL,LI,PRE,STRONG,EM,BR,HR)

   Graphic Elements

     HR: Horizontal ruler
   
   """


import string,UserList
import operator,re
import Globals,pdfutilities
from pdflib import PDFObject

stringwidth=pdfutilities.stringwidth


def escape(txt,replace=string.replace):
    """escape(txt)

    escapes banned PDF characters"""

    ## ugly but faster!
    return replace(replace(txt,'(','\('),')','\)')
    
strip_front=re.compile('^[\n ]*\n',re.S)
strip_back=re.compile('\n[\n ]*$',re.S)

class Element:
    '''Element mixin class'''

    def font_family(self):
        """Replaces font-family's actual name to the one defined in
        the document resources"""
        
        return Globals.font_map[self['font_family']]


class Phrase(PDFObject,Element):
    """A Phrase which can include only a single letter has the
    following attributes:

    - font_family

    - font_size
    
    - letter_spacing

    - leader (distance in points between the baseline of phrase and the
      baseline of the previous line)

    - word_spacing
    """

    element_type='Phrase'
    
    def __init__(self,text,props=None):
        """text_properties is a dictionary of the above listed
        attributes"""
        self.text=text
        PDFObject.__init__(self,props)

        
    def get_bbox(self):
        """get_bbox(self)

        returns bounding box of text"""

        ## stringwidth returns the  total width of the text based on
        ## the afm files. From Andy Robinson
        
        font_family=self['font_family']
        w=stringwidth(self.text,font_family)
        width=int(w*0.001*self['font_size']+len(self.text)*self['letter_spacing']+0.5)
        height=self['leader'] ## for future support of subscripts
        return width,height

    def get_spacewidth(self):
        font_name=self['font_family']
        w=stringwidth(' ',font_name)
        return int(w*0.001*self['font_size']+self['word_spacing']+0.5)

    def to_PDF(self,plain=0):
        """ Output phrase""" 
        if plain:
            return escape(self.text)
        else:
            return '%s %s Tf (%s) Tj'%(self.font_family(),
                                  self['font_size'],escape(self.text))


class Phrases(UserList.UserList):
    '''Holds a list of Phrases with common properties'''

    def to_PDF(self,mode='normal'):
        w=self[0]
        txt=string.join(map(lambda a:a.text,self.data),' ')
        if mode=='normal':
            out='%s %i Tf (%s) Tj'%(w.font_family(),
                                    w['font_size'],escape(txt))
        elif mode=='plain':
            out='(%s) Tj'%escape(txt)
        elif mode=='leading_space':
            t=w.text
            if len(t)==1 and t in '.,:;?!':
                out='%s %i Tf (%s) Tj'%(w.font_family(),
                                         w['font_size'],escape(txt))
            else:
                out='%s %i Tf ( %s) Tj'%(w.font_family(),
                                         w['font_size'],escape(txt))
        return out


class Paragraph(PDFObject,Element):
    
    """
    Main text element tag. the majority of the other ags will be
    build around this class.

    A paragraph is represented internally by a list of two-element
    tuples. The first element in each tuple is a four element tuple
    that contains the width,height,word spacing and x_offset of the
    line. The second element is a list of Phrase instances.
    """

    element_type='P'

    def __init__(self,props=None):
        self.phrases=[]
        PDFObject.__init__(self,props)
        self['counter_set_flag']=0
        
    def add_text(self,text,style='NORMAL'):
        """add_text(self,text,style='NORMAL')

        Constructs paragraph from a block of text"""

        text=string.replace(text,'\n',' ')
        text=string.replace(text,'\r',' ')
        text=string.split(text)
        phrases=[]

        ## yes yes I could have the if in the loop but it is faster
        ## like this ...

        ## A hack as the bold and italic styles should apply on the
        ## current font_family instead of hard-wiring the font family
        ## as well. But it is easy this way!

        if style=='NORMAL':
            for word in text:
                w=Phrase(word)
                phrases.append(w)
        elif style=='CODE':
            for word in text:
                w=Phrase(word)
                w['font_family']='Courier'
                phrases.append(w)
        elif style=='STRONG':
            for word in text:
                w=Phrase(word)
                w['font_family']='Times-Italic'
                phrases.append(w)
        elif style=='EM':
            for word in text:
                w=Phrase(word)
                w['font_family']='Times-Bold'
                phrases.append(w)
        else:
            for word in text:
                w=Phrase(word)
                phrases.append(w)
                
        self.add_phrases(phrases)

            
    def add_phrases(self,phrases):
        """add_phrases(self,phrases)

        Constructs a paragraph from a list of phrases"""

        for phrase in phrases:
            phrase.set_parentProperties(self)
            self.phrases.append(phrase)

    def get_text(self):
        """get_text(self)

        returns paragraphs contents as a plain text string"""
        txt=''
        for phrase in self.phrases:
            txt='%s %s'%(txt,phrase.text)
        return string.strip(txt)

    def _add_lines(self,lines):
        """_add_lines(self,lines)

        Constructs a paragraph from a list of lines. See method format
        for definition of lines"""

        phrases=[]
        for line in lines:
            phrases=phrases+line[1]
        self.add_phrases(phrases)

    def set_bullet(self,bullet):
        
        """set_bullet(self,bullet)

        sets bullet"""
        bullet.set_parentProperties(self)
        self.bullet=bullet

    def format(self):
        """format(self)

        to format a paragraph the following attributes should be
        accessible: 

        - cursor_x: the x coordinate of the cursor on the current page

        - cursor_y: the y coordinate of the cursor on the current
          page. They will be generally provided from the page instance
          that contains the paragraph

        - text_area_width: paragraph should format its contents to fit
          this width (unless left_offset is set)

        - P_left_offset: number of points the paragraph will be
          displaced from the left margin

        - alignment: ragged-right, justified, centered, right

        - level: paragraph level. Allows proper positioning inside
          lists

        - level_offset: level point offset

        - tw_range: word spacing shrink and stretch factors. Needed
          for justifiacation.

        - indent: optional first line indent

        - P_leader: the distance between the baseline of the last line
          of the paragraph and the following one.

        """

        self['par_cursor_y']=self['cursor_y']-self['P_top_dy']
        loff=self['P_left_offset']+self['level_offset']*self['level']
        self['par_text_area_width']=self['text_area_width']-loff
        self['par_cursor_x']=self['cursor_x']+loff
        tw_shrink,tw_stretch=self['tw_range']
        font_name=self['font_family']
        w=stringwidth(' ',font_name)
        space_width=int(w*0.001*self['font_size']+self['word_spacing']+0.5)
        
        par_h=self['P_top_dy']
        lines=[]
        line=[]
        line_break=0
        
        line_w=self['indent']-space_width
        x_offset=self['indent']  ## take care of first line  
        line_h=self.phrases[0].get_bbox()[1]

        ## This loop should be optimized
        par_text_area_width=self['par_text_area_width']
        for word in self.phrases:
            if word.element_type=='BR':
                line_break=1
            w,h=word.get_bbox()
            ##space=word.get_spacewidth()+self['word_spacing']
            line_w=line_w+w+space_width
            if line_w<=par_text_area_width and line_break==0:
                line.append(word)
                line_h=max(line_h,h) ## In case we have sub-super-scripts
            else:
                ##if line_break==1:print word.text
                line_break=0
                line_w=line_w-(w+space_width)
                ## format line
                if self['alignment']=='ragged-right':
                    fmt_line=((line_w,line_h,-1,x_offset),line)
                
                elif self['alignment']=='right':
                    x_offset=x_offset+par_text_area_width-line_w
                    fmt_line=((line_w,line_h,-1,x_offset),line)
                
                elif self['alignment']=='center':
                    x_offset=(x_offset+par_text_area_width-line_w)/2.0
                    fmt_line=((line_w,line_h,-1,x_offset),line)
                lines.append(fmt_line)
                par_h=par_h+line_h
                x_offset=0
                line=[word]
                line_w=w
                line_h=h
            
        if line:
            if self['alignment']=='ragged-right':
                fmt_line=((line_w,line_h,-1,x_offset),line)
                
            elif self['alignment']=='right':
                x_offset=x_offset+par_text_area_width-line_w
                fmt_line=((line_w,line_h,-1,x_offset),line)
                
            elif self['alignment']=='center':
                x_offset=(x_offset+par_text_area_width-line_w)/2.0
                fmt_line=((line_w,line_h,-1,x_offset),line)
            lines.append(fmt_line)
            par_h=par_h+line_h
            
        self.lines=lines
        self.par_h=par_h+self['P_bottom_dy']

    def get_bbox(self):
        """get_bbox(self)

        returns bounding box of paragraph. Method format must be
        called before this method"""

        return self['par_text_area_width'],self.par_h


    def more(self):

        """Returns a paragraph, and possibly a second paragraph
        containing any text that could not fit in the remaining page
        space.

        A simple formatting rule is followed. A paragraph will be
        split only if there are at least 'break_widows' lines on the
        current page and 'break_orphans' on the following page.

        The following attributes must exist in the properties
        dictionary:
        
        - available_space: the vertical space left in the current page

        - break_widows: minimum number of widow lines aloowed

        - break_orphans: minimum number of orphan lines allowed

        Method format must be called before calling more"""

        page_space=self['available_space']
        break_widows=self['break_widows']
        break_orphans=self['break_orphans']
        
        if self.par_h<=page_space:
            return self,None
        elif len(self.lines)<break_widows+break_orphans:
            return None,self
        else:
            lines=[]
            par_h=0
            count=0
            no_lines=len(self.lines)
            while par_h<=page_space:
                if count>=no_lines:break
                par_h=par_h+self.lines[count][0][1]
                count=count+1
                
            if count<break_widows+1:
                return None,self
            
            elif len(self.lines)-count<break_orphans-1:
                par_props=self.properties.parent_properties
                b=Paragraph(self.properties.data)
                b['indent']=0
                b._add_lines(self.lines[-break_widows:])
                b.properties.parent_properties=par_props
                b.format()
                self.phrases=[]
                self._add_lines(self.lines[:-break_widows])
                self.format()
                return self,b

            else:
                par_props=self.properties.parent_properties
                b=Paragraph(self.properties.data)
                b['indent']=0
                b._add_lines(self.lines[count-1:])
                b.properties.parent_properties=par_props
                b.format()
                self.phrases=[]
                self._add_lines(self.lines[:count-1])
                self.format()
                return self,b

    def to_PDF(self):
        """to_PDF(self)

        Returns a PDF representation fo the paragraph. Method format
        must be called before calling this method."""
        
        ## Lets not worry too much about optimization right now

        if hasattr(self,'bullet'):
            bl_x=self['par_cursor_x']-self['bullet_x_offset']
            bl_y=self['par_cursor_y']-self['leader']
            out=['BT',
                 '1 0 0 1 %i %i Tm'%(bl_x,bl_y),
                 self.bullet.to_PDF()]
        else:
            out=['BT']
        ## Set transformation matrix to absolute position of paragraph
        tm="1 0 0 1 %(par_cursor_x)i %(par_cursor_y)i Tm"%self
        out.append("1 0 0 1 %(par_cursor_x)i %(par_cursor_y)i Tm"%self)        
        font_family=self.font_family()
        out.append("""%s %s Tf"""% (font_family,self['font_size']))

        ## Don't worry about optimization yet
        props=self.lines[0][1][0].properties ## props of first word

        old_x_offset=0
        for line in self.lines:
            w,leader,sp_w,dx=line[0]
            x_offset=dx-old_x_offset
            old_x_offset=dx
            out.append('%i -%i Td'%(x_offset,leader))
            if sp_w>0:
                out.append('%i Tw'%sp_w)

            groups=[] ##groups phrases with common properties
            phrases=Phrases()
            for word in line[1]:
                if word.element_type=='BR':continue
                
                if word.properties==props:
                    phrases.append(word)
                else:
                    if phrases:
                        groups.append(phrases)
                    phrases=Phrases([word])
                    props=word.properties
                    font_family=word['font_family']
            if phrases:
               groups.append(phrases)
               
            if len(groups)>1:
                out.append(groups[0].to_PDF())
                for group in groups[1:]:
                    out.append(group.to_PDF('leading_space'))
            else:
                ## To be optimized
                g_font=groups[0][0].font_family()
                if g_font==font_family:
                    out.append(groups[0].to_PDF('plain'))
                else:
                    out.append(groups[0].to_PDF()) ## to be optimized
                    font_family=g_font

        out.append('ET')
        return string.join(out,'\n')

P=Paragraph

class PRE(PDFObject,Element):
    """Example text. Careful, no line wrapping"""

    element_type='PRE'
    
    def __init__(self,props=None):
        self.phrases=[]
        PDFObject.__init__(self,props)    

    def add_text(self,text):
        """add_text(self,text)

        Constructs paragraph from a block of text"""

        text=string.replace(text,'\n\r','\n')
        x=strip_front.search(text)
        y=strip_back.search(text)
        if x:
            x=x.end()
        else:
            x=0
        if y:
            y=y.start()
        else:
            y=len(text)
        text=string.split(text[x:y],'\n')        

        phrases=[]
        for line in text:
            ph=Phrase(line)
            phrases.append(ph)
        self.add_phrases(phrases)

        
    def add_phrases(self,phrases):
        """add_phrases(self,phrases)

        Constructs a paragraph from a list of phrases"""

        for phrase in phrases:
            phrase.set_parentProperties(self)
            self.phrases.append(phrase)


    def set_bullet(self,bullet):
        
        """set_bullet(self,bullet)

        sets bullet"""
        bullet.set_parentProperties(self)
        self.bullet=bullet
        
    def format(self):
        """format(self)

        to format a PRE section the following attributes should be
        accessible: 

        - PRE_font_family: Font family for PRE sections.

        - cursor_x: the x coordinate of the cursor on the current page

        - cursor_y: the y coordinate of the cursor on the current
          page. They will be generally provided from the page instance
          that contains the paragraph

        - text_area_width: paragraph should format its contents to fit
          this width (unless left_offset is set)

        - PRE_left_offset: number of points the paragraph will be
          displaced from the left margin

        - level: paragraph level. Allows proper positioning inside
          lists

        - level_offset: level point offset

        - PRE_leader: the distance between the baseline of the last line
          of the paragraph and the following one.

        """
        self['font_family']=self['PRE_font_family']
        self['par_cursor_y']=self['cursor_y']-self['PRE_top_dy']
        loff=self['PRE_left_offset']+self['level_offset']*self['level']
        self['par_cursor_x']=self['cursor_x']+loff
        self.par_h=self['leader']*len(self.phrases)+\
                    self['PRE_top_dy']+self['PRE_bottom_dy']
        width=max(map(lambda a:a.get_bbox()[0],self.phrases))
        self['par_textarea_width']=width

    def get_bbox(self):
        """get_bbox(self)

        returns bounding box of paragraph. Method format must be
        called before this method"""

        return self['par_textarea_width'],self.par_h


    def more(self):

        """Returns a paragraph, and possibly a second paragraph
        containing any text that could not fit in the remaining page
        space.

        A simple formatting rule is followed. A paragraph will be
        split only if there are at least 'break_widows' lines on the
        current page and 'break_orphans' on the following page.

        The following attributes must exist in the properties
        dictionary:
        
        - available_space: the vertical space left in the current page

        - break_widows: minimum number of widow lines aloowed

        - break_orphans: minimum number of orphan lines allowed

        Method format must be called before calling more"""

        pagespace=self['available_space']
        break_widows=self['break_widows']
        break_orphans=self['break_orphans']

        no_lines=len(self.phrases)
        lines_here=int(divmod(pagespace,self['leader'])[0])
        lines_other=no_lines-lines_here

        if self.par_h<=pagespace:
            return self,None
        elif no_lines<break_widows+break_orphans:
            return None,self
        else:
                
            if lines_here<break_widows:
                return None,self
            
            elif lines_other>=break_orphans:
                par_props=self.properties.parent_properties
                b=self.__class__(self.properties.data)
                b.add_phrases(self.phrases[lines_here:])
                b.properties.parent_properties=par_props
                del self.phrases[lines_here:]
                self.format()
                return self,b

            else:
                par_props=self.properties.parent_properties
                b=self.__class__(self.properties.data)
                b.add_phrases(self.phrases[-break_orphans:])
                b.properties.parent_properties=par_props
                del self.phrases[-break_orphans:]
                self.format()
                return self,b

    def to_PDF(self):
        """to_PDF(self)

        Returns a PDF representation of the example text. Method format
        must be called before calling this method."""

        if hasattr(self,'bullet'):
            bl_x=self['par_cursor_x']-self['bullet_x_offset']
            bl_y=self['par_cursor_y']-self['leader']
            out=['BT',
                 '1 0 0 1 %i %i Tm'%(bl_x,bl_y),
                 self.bullet.to_PDF()]
        else:
            out=['BT']
            
        ## set the trsnformation matrix so all elements can be
        ## specified relative to (0,0)

        leader=self['leader']
        out.append("""1 0 0 1 %(par_cursor_x)i %(par_cursor_y)i Tm"""% self)
        out.append("""%s %s Tf"""% (self.font_family(),self['font_size']))
        out.append("""%i TL"""% leader)

        for phrase in self.phrases:
            out.append("""(%s) '"""%phrase.to_PDF(plain=1))
        out.append('ET')

        return string.join(out,'\n')


class UL(PDFObject):
    """Unordered List.

    LI: takes a list of elements that form an LI group.
    LI elements can be of type P,PRE,UL,OL and DL.
    
    If you want to use UL without bullets just set the 'bullet_char'
    to empty space.
    """

    element_type='UL'
    
    def __init__(self,props=None):
        self.list_items=[]
        self.level=1
        PDFObject.__init__(self,props)
        
    def LI(self,elements):
        if type(elements)!=type([]):
            raise TypeError,'LI argument must be a list'
        self.list_items.append(elements)

        
    def get_elements(self):
        out=[]
        for list_item in self.list_items:
            is_first=1
            for element in list_item:
                if element.element_type in ['P','PRE']:
                    element['level']=self.level
                    element['P_bottom_dy']=self['LI_bottom_dy']
                    element['P_top_dy']=self['LI_top_dy']
                    element['indent']=0

                    if is_first:
                        ##make bullet
                        bl=Phrase(self['bullet_char'])
                        bl['font_family']=self['bullet_font_family']
                        element.set_bullet(bl)
                        is_first=0
                        
                elif element.element_type in ['UL','OL','DL']:
                    element.level=self.level+1
                out.append(element)
                
        return out
            


class OL(PDFObject):
    '''Ordered List.'''

    element_type='OL'
    
    def __init__(self,props=None):
        self.level=1
        self.list_no=pdfutilities.Counter()
        self.list_items=[]
        PDFObject.__init__(self,props)

        
    def LI(self,elements):
        if type(elements)!=type([]):
            raise TypeError,'LI argument must be a list'
        self.list_items.append(elements)

        
    def get_elements(self):
        out=[]
        for list_item in self.list_items:
            is_first=1
            for element in list_item:
                if element.element_type in ['P','PRE']:
                    element['level']=self.level
                    element['P_bottom_dy']=self['LI_bottom_dy']
                    element['P_top_dy']=self['LI_top_dy']
                    element['indent']=0
                    if is_first:
                        ##make bullet
                        no='%i.'%self.list_no.incr()
                        bl=Phrase(no)
                        bl['font_family']=self['OL_font_family']
                        element.set_bullet(bl)
                        is_first=0
                elif element.element_type in ['UL','OL','DL']:
                    element.level=self.level+1
                out.append(element)
        return out

class H:
    '''Header base class'''

    def set_counter(self,counter):
        co=Phrase(counter)
        co.set_parentProperties(self)
        self.phrases.insert(0,co)
        
    def format(self):
        self['P_left_offset']=self['%s_left_offset'%self.element_type]
        self['level']=0
        self['P_bottom_dy']=self['%s_bottom_dy'%self.element_type]
        self['P_top_dy']=self['%s_top_dy'%self.element_type]
        self['font_family']=self['%s_font_family'%self.element_type]
        self['font_size']=self['%s_font_size'%self.element_type]
        self['indent']=0
        self['bullet_x_offset']=self['H_bullet_x_offset']
        Paragraph.format(self)

class H1(H,Paragraph):
    '''Aha! H has to come before Paragraph in the base classes'''
    element_type='H1'

class H2(H,Paragraph):
    element_type='H2'

class H3(H,Paragraph):
    element_type='H3'

class H4(H,Paragraph):
    '''Aha! H has to come before Paragraph in the base classes'''
    element_type='H4'

class H5(H,Paragraph):
    element_type='H5'

class H6(H,Paragraph):
    element_type='H6'

class Line(PDFObject):
    '''Draws a line from x1,y1 to x2,y2 with thickness line_thickness'''
    element_type='Line'

    def format(self):
        pass

    def more(self):
        return self,None

    def get_bbox(self):
        '''get_bbox(self)

        no bounding box info neccessary'''
         
        return 0,0
    
    def to_PDF(self):
        out=['%i w'%self['line_thickness']]
        out.append('%i %i m'%(self['x1'],self['y1']))
        out.append('%i %i l'%(self['x2'],self['y2']))
        out.append('S')
        return string.join(out,'\n')


class HR(Line):
    element_type='HR'
    
    def format(self):
        h=self['line_thickness']=self['HR_thickness']
        self.height=h+self['HR_bottom_dy']+self['HR_top_dy']
        text_width=self['text_area_width']
        w=self['HR_width']
        if type(w)==type(''):
            percentage=string.atof(string.strip(w)[:-1])/100
            self.width=text_width*percentage
        else:
            self.width=w

        alignment=self['alignment']
        if alignment=='ragged-right':
            x_offset=0
        elif alignment=='center':
            x_offset=(text_width-self.width)/2.0
        elif alignment=='right':
            x_offset=text_width-self.width

        x1=self['x1']=self['left_margin']+x_offset
        self['x2']=x1+self.width
        self['y1']=self['y2']=self['cursor_y']-self['HR_top_dy']
        
    def get_bbox(self):
        return self.width,self.height

    
class Header(Element,PDFObject):
    '''Sets page Headers'''
    
    element_type='Header'

    def get_bbox(self):

        ## headers are absolutely positioned
        ## we do not want the page instance to adjust available space.
        
        return 0,0


    def more(self):
        return self,None
    
    def add_page_no(self):
        self.add_page_no_flag=1

    def format(self):

        ## Add header text if available
        if self['Header_text']:
            self.text=self['Header_text']

        ## Add page no if add_page_no_flag is set
        if hasattr(self,'add_page_no_flag'):
            if self.text:
                self.text='%s %s'%(self.text,string.rjust
                                   (self['current_page_number'],4))
            else:
                self.text=self['current_page_number']


        font_size=self['font_size']=self['Header_font_size']
        font_family=self['font_family']=self['Header_font_family']
        w=stringwidth(self.text,font_family)
        width=int(w*0.001*font_size+len(self.text)*self['letter_spacing']+0.5)
           
        position=self['position']
        left_margin=self['left_margin']
        right_margin=self['right_margin']
        top_margin=self['top_margin']
        bottom_margin=self['bottom_margin']
        page_width=self['page_width']
        page_height=self['page_height']
        if position=='TR':
            self['x']=page_width-(right_margin+width)
            self['y']=page_height-top_margin+self['Header_top_y']
        elif position=='TL':
            self['x']=left_margin
            self['y']=page_height-top_margin+self['Header_top_y']
        elif position=='BL':
            self['x']=left_margin
            self['y']=bottom_margin-self['Header_bottom_y']-font_size
        elif position=='BR':
            self['x']=page_width-(right_margin+width)
            self['y']=bottom_margin-self['Header_bottom_y']-font_size
            
    def to_PDF(self):
        """ Output header"""
        out=['BT']
        out.append('1 0 0 1 %i %i Tm'%(self['x'],self['y']))
        out.append('%s %s Tf (%s) Tj'%(self.font_family(),
                       self['font_size'],escape(self.text)))
        out.append('ET')

        return string.join(out,'\n')
    
class HeaderLine(Line):
        
    element_type='HeaderLine'
        
    def format(self):
        self['line_thickness']=self['Header_line_thickness']
        position=self['position']
        left_margin=self['left_margin']
        right_margin=self['right_margin']
        top_margin=self['top_margin']
        bottom_margin=self['bottom_margin']
        page_width=self['page_width']
        page_height=self['page_height']
        y_offset=self['Header_line_y_offset']
        self['x1']=left_margin
        self['x2']=page_width-right_margin
        if position=='T':
            self['y1']=self['y2']=page_height-top_margin+y_offset
        elif position=='B':
            self['y1']=self['y2']=bottom_margin-y_offset

class BR(PDFObject):
    """BR acts like \linebreak in Latex and not like BR in
    HTML. Adding many BR elements will not add blanklines"""
    
    element_type='BR'
    def get_bbox(self):
        return 0,0
    def to_PDF(self):
        pass
