# -*- Mode: Python; tab-width: 4 -*-
#
#	$Id: pdflib.py,v 1.4 1999/06/17 17:39:30 pavlos Exp $
#	Author: Pavlos Christoforou <pavlos@gaaros.com>
#
# ======================================================================
#
# This software, its author and the author's donkey are all certified
# as open source. Use and abuse us in any way you see fit.
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY GAAROS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL GAAROS OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
########################################################################

__version__='$Revision: 1.4 $'[11:-2]


"""\n
Zpdf

  Native python library for generating Portable Document Format (PDF)
  documents. Provided mainly in support of StruturedText and is
  therefore limited to text and simple graphics output. However it
  provides a nice class interface to the library.

  The main structure in a PDF document is a Document which contains a
  number of elements defined in the elements.py module. The pages are
  collected in a hierarchical manner under a root object and can share
  common resources through shared dictionaries. The PDF language is,
  like postscript a general page description language, and provides no
  tools for higher level document formatting. The object model of this
  module was designed for simple applications and facilitates easy
  document formatting. Document formating is a complex and difficult
  subject and thus users who demand a complete system should probably
  look at Lout or TeX.

  The library was inspired by similar efforts, mainly by Andy Robinson
  and David Ascher.   

$Id: pdflib.py,v 1.4 1999/06/17 17:39:30 pavlos Exp $"""

__version__='$Revision: 1.4 $'[11:-2]

__author__='Pavlos Christoforou'

import Globals
import UserDict
import string,time,pdfutilities
from pdfutilities import Counter
from styles import default
import copy

OID=Globals.OID

H_type=['H1','H2','H3','H4','H5','H6']

class AcquisitionDict(UserDict.UserDict):
    """Simple acquisition dict. See also PDFObject class.
    Accesses are first checked against the local properties
    dictionary. If the requested key is not found, an attempt is made
    to retrieve the request from the parent properties dictionary."""

    def __init__(self,props=None):
        self.parent_properties={}
        UserDict.UserDict.__init__(self)
        if props:
            for k,v in props.items():
                self[k]=v

    def __getitem__(self,key):
        if self.data.has_key(key):
            return self.data[key]
        try:
            value=self.parent_properties[key]
        except KeyError:
            raise Globals.PropertyError(key)
        return value
    
    def __setitem__(self,key,value):

        ## Some tag elements unavoidably affect other related tags.
        ## for instance changin font size should change font leader too
        ## here is a good place to intercept any changes and modify
        ## any other relevant tags
        
        if key not in ['font_family','font_size']: ## for speed
            self.data[key]=value
        elif key=='font_family':
            if Globals.font_map.has_key(value):
                self.data[key]=value
            else:
                Globals.Font(value)
                self.data[key]=value
        elif key=='font_size':
            self.data[key]=value
            ## crude leader selection
            self.data['leader']=value*1.2

class PDFObject:
    ''' PDFObject mix-in class

    AcqusitionDict methods could have been added here but then
    circular reference problems might arise. In this way we do not
    need to keep a reference to the parent object, just to its
    properties dict.'''
    
    def __init__(self,props=None):
        self.properties=AcquisitionDict(props)

    def __getitem__(self,key):
        return self.properties[key]
    
    def __setitem__(self,key,value):
        self.properties[key]=value

    def has_key(self,key):
        return self.properties.has_key(key)
    
    def set_parentProperties(self,parent):
        self.properties.parent_properties=parent.properties


class Outlines:
    'Not implemented yet'
    def __init__(self):
        self.OID=OID.incr()
        
    def to_PDF(self):
        out=['%i 0 obj'%self.OID,
             '<<',
             '/Type /Outlines',
             '/Count 0',
             '>>',
             'endobj']
        return string.join(out,'\n')


class Info(PDFObject):
    'Minimal info element'
    
    def __init__(self):
        self.OID=OID.incr()
        PDFObject.__init__(self,props=None)
        
    def to_PDF(self):
        out=['%i 0 obj'%self.OID,
             '<</Title (%s)'%self['title'],
             ## '/Creator (%s)'%self['creator'],
             '/Author (%s)'%self['author'],
             '/CreationDate (D:%s)'%time.strftime('%Y%m%d%H%M',
                                                  (time.gmtime(time.time()))),
             '/Creator (Zpdf v0.1)',
             '>>',
             'endobj']

        return string.join(out,'\n')




class Document(PDFObject):
    """The main class.

       A container of pages. A simple model is followed where all
       pages are contained in one pages pdf object. Individual pages
       contain no resources but instead inherit a common resource
       object from the pages object."""

    def __init__(self,props=default):
        PDFObject.__init__(self,props)
        self.OID=OID.incr()
        self['pages_OID']=OID.incr()
        self.pages=[]
        self.page_counter=Counter()
        self.elements=[]
        self.common_elements=[]
        
    def initialize(self):
        self.outlines=Outlines()
        self.info=Info()
        self.info.set_parentProperties(self)
        
    def add_elements(self,elements):
        '''add_elements(self,elements)

        add all elements of document'''

        ## remove empty elements

        good_elements=[]
        for el in elements:
            if hasattr(el,'phrases'):
                if el.phrases:
                    good_elements.append(el)
            else:
                good_elements.append(el)
                
        self.elements=self.elements+good_elements

    def sh(self):
        '''Debugging aid'''
        print Globals.resources_map
        print Globals.font_list

    def add_commonElements(self,elements):
        """add_commonElements(self,elements)

        elements that appear on all pages"""
        self.common_elements=self.common_elements+elements


    def set_style(self,style):
        for k,v in style.items():
            self[k]=v
        
    def format(self):
        """format(self)

        iterates through elements and fills up pages"""

        h1_counter=Counter()
        h2_counter=Counter()
        h3_counter=Counter()

        self.elements_stack=pdfutilities.CrazyStack(self.elements[:])
        self.current_page=self.new_page()
        page_list=[]

        while self.elements_stack:
            element=self.elements_stack.pop()
            el_type=element.element_type
            ##print element.element_type,len(self.elements_stack)
            if el_type in ['UL','OL','DL']:
                element.set_parentProperties(self.current_page)
                self.elements_stack.push(element.get_elements())
                element=self.elements_stack.pop()

            ## Set a counter for elements H1 to H3
            elif el_type in ['H1','H2','H3']:
                if not element['counter_set_flag']:
                    if el_type=='H1':
                        h1_counter.incr()
                        h2_counter.reset()
                        h3_counter.reset()
                        if self['H1_numbering']:
                            element.set_counter('%s.0'%str(h1_counter))
                    
                    elif el_type=='H2':
                        h2_counter.incr()
                        h3_counter.reset()
                        if self['H2_numbering']:
                            element.set_counter('%s.%s'%(str(h1_counter),
                                                 str(h2_counter)))
                    elif el_type=='H3':
                        h3_counter.incr()
                        if self['H3_numbering']:
                            element.set_counter('%s.%s.%s'%(
                            str(h1_counter),str(h2_counter),str(h3_counter)))
                else:
                    element['counter_set_flag']=0
                    
                if self.is_first_section:
                    ## Set Top Header
                    if el_type==self['Header_section']:
                        self['Header_text']=element.get_text()
                        self.is_first_section=0

            element.set_parentProperties(self.current_page)
            element.format()

            here,other=element.more()
            if not other:
                page_list.append(here)
                w,h=here.get_bbox()
                self['cursor_y']=self['cursor_y']-h
                self['available_space']=self['available_space']-h

            else:
                ## I hope nobody writes a title that spans more than 8
                ## lines!
                
                if here:
                    page_list.append(here)
                    here['counter_set_flag']=1
                                    
                ## If last element in page is a title then move it to
                ## next page
                else:
                    if page_list[-1].element_type in ['H1','H2',
                                                      'H3','H4','H5','H6']:
                        first=page_list[-1]
                        ## set counter flag so it won't be updated again
                        first['counter_set_flag']=1
                        self.elements_stack.push([first,other])
                        del page_list[-1]
                    else:
                        self.elements_stack.push(other)

                ##add common elements at end of page
                for el in self.common_elements:
                    el_copy=copy.deepcopy(el)
                    el_copy.set_parentProperties(self.current_page)
                    el_copy.format()
                    page_list.append(el_copy)

                self.current_page.add_elements(page_list)
                self.current_page=self.new_page()
                page_list=[]

        if page_list:
            for el in self.common_elements:
                el_copy=copy.deepcopy(el)
                el_copy.set_parentProperties(self.current_page)
                el_copy.format()
                page_list.append(el_copy)
            self.current_page.add_elements(page_list)
            
    def new_page(self):
        page=Page()
        self.page_counter.incr()
        page['current_page_number']=str(self.page_counter)
        page.set_parentProperties(self)
        self.pages.append(page)
        self['cursor_x']=self['left_margin']
        self['cursor_y']=self['page_height']-self['top_margin']
        self['available_space']=self['text_area_height']
        self.is_first_section=1
        return page

    def to_PDF(self):
        out='%PDF-1.0\n%\n'
        kids=''
        for page in self.pages:
            kids=kids+'%i 0 R '%page.OID

        catalog=['%i 0 obj'%self.OID,
                 '<<',
                 '/Type /Catalog',
                 '/Pages %i 0 R'%self['pages_OID'],
                 '/Outlines %i 0 R'%self.outlines.OID,
                 '>>',
                 'endobj']
        
        pages=['%i 0 obj'%self['pages_OID'],
               '<<',
               '/Type /Pages',
               '/Count %i'%len(self.pages),
               '/Kids [%s]'%kids,
               '/Resources <<',
               '             /ProcSet [/PDF /Text]',
               '             /Font',
               pdfutilities.dict_to_pdfdict(Globals.resources_map),
               '           >>',
               '>>',
               'endobj']
        
        xrefs={}

        
        xrefs[self.OID]=len(out)
        out=out+string.join(catalog,'\n')+'\n'

        xrefs[self['pages_OID']]=len(out)
        out=out+string.join(pages,'\n')+'\n'
        
        objects=[(self.outlines.OID,self.outlines.to_PDF()),
                 (self.info.OID,self.info.to_PDF())]

        for page in self.pages:
            p,s=page.to_PDF()
            objects.append((page.OID,p))
            objects.append((page.contents_OID,s))

        for font in Globals.font_list:
            objects.append((font.OID,font.to_PDF()))

        for obj in objects:
            xrefs[obj[0]]=len(out)
            out=out+obj[1]+'\n'

        no_objs=len(xrefs)+1
        obj_oids=xrefs.keys()
        obj_oids.sort()
        xrefs_xref=len(out)
        out=out+'xref\n%i %i\n0000000000 65535 f\n'%(0,no_objs)
        for obj_oid in obj_oids:
            out=out+'%0.10d 00000 n\n'%xrefs[obj_oid]
        out=out+'trailer\n'
        out=out+'<< /Size %i /Root %i 0 R /Info %i 0 R>>\n'%(no_objs,
                                            self.OID,self.info.OID)
        out=out+'startxref\n'
        out=out+'%i\n'%xrefs_xref
        out=out+'%%EOF'
        return out

    
class Page(PDFObject):
    """Implements a Page object. Provided for class Document support"""

    def __init__(self,props=None):
        PDFObject.__init__(self,props)
        self.OID=OID.incr()
        self.contents_OID=OID.incr()
        self.elements=[]

    def add_elements(self,elements):
        """add_elements(self,elements)"""

        for element in elements:
            element.set_parentProperties(self)
            self.elements.append(element)
        ## print 'Added %i elements'%len(elements)
     
    def to_PDF(self):
        p_out=['%i 0 obj'%self.OID,
               '<<',
               '/Type /Page',
               '/Parent %i 0 R'%self['pages_OID'],
               '/Resources null',
               '/MediaBox [0 0 %(page_width)i %(page_height)i]'%self, 
               '/Contents %i 0 R'%self.contents_OID, 
               '>>',
               'endobj']

        stream=[]
        for element in self.elements:
            stream.append(element.to_PDF())
        stream=string.join(stream,'\n')
        length=len(stream)+1

        s_out=['%i 0 obj'%self.contents_OID,
                 '<< /Length %i >>'%length,
                 'stream',
                 '\r%s'%stream,
                 'endstream',
                 'endobj']
                 
        return [string.join(p_out,'\n'),string.join(s_out,'\n')]

